# Plot.py -- Pamhyr
# Copyright (C) 2023-2024  INRAE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

import logging

from math import dist, sqrt

from tools import timer, trace
from View.Tools.PamhyrPlot import PamhyrPlot

from PyQt5.QtCore import (
    Qt, QCoreApplication, QItemSelectionModel
)

from PyQt5.QtWidgets import QApplication

_translate = QCoreApplication.translate

logger = logging.getLogger()


class Plot(PamhyrPlot):
    def __init__(self, canvas=None, trad=None, data=None, toolbar=None,
                 table=None, parent=None):
        super(Plot, self).__init__(
            canvas=canvas,
            trad=trad,
            data=data,
            toolbar=toolbar,
            parent=parent
        )

        self._table = table
        self._z_note = None
        self._z_line = None
        self._z_fill_between = None

        self.line_xy = []
        self.line_gl = []

        self.label_x = self._trad["unit_kp"]
        self.label_y = self._trad["unit_height"]

        self.before_plot_selected = None
        self.plot_selected = None
        self.after_plot_selected = None

        self._isometric_axis = False

        self.hl_points = []
        self.highlight = (
            [],                 # Points list to highlight
            None                # Hydrolic values (z, wet_area,
                                # wet_preimeter, water_width)
        )

    def onpick(self, event):
        if event.mouseevent.inaxes != self.canvas.axes:
            return

        modifiers = QApplication.keyboardModifiers()
        if modifiers != Qt.ControlModifier:
            return

        _, hyd = self.highlight

        ind, point = self._closer_point(event)

        self.highlight = ([point], hyd)
        self._select_in_table(ind)

        self.update()
        return

    def onclick(self, event):
        if event.inaxes != self.canvas.axes:
            return

        modifiers = QApplication.keyboardModifiers()
        if modifiers != Qt.ShiftModifier:
            return

        points, _ = self.highlight

        z = self._get_z_from_click(event)
        if z < self.data.z_min():
            return

        a, p, w = self._compute_hydrolics(z)

        logger.debug(f"{z, a, p, w}")

        self.highlight = (points, (z, a, p, w))

        self.update()
        return

    def select_points_from_indexes(self, indexes):
        data = self.data
        _, hyd = self.highlight

        points = list(
            map(
                lambda e: e[1],
                filter(
                    lambda e: e[0] in indexes,
                    enumerate(
                        zip(data.get_station(), data.z())
                    )
                )
            )
        )

        self.highlight = (points, hyd)
        self.update()

    def _select_in_table(self, ind):
        if self._table is not None:
            self._table.blockSignals(True)
            self._table.setFocus()

            selection = self._table.selectionModel()
            index = self._table.model().index(ind, 0)
            selection.select(
                index,
                QItemSelectionModel.Rows |
                QItemSelectionModel.ClearAndSelect |
                QItemSelectionModel.Select
            )

            self._table.scrollTo(index)
            self._table.blockSignals(False)

    def _closer_point(self, event):
        points_ind = event.ind
        axes = self.canvas.axes
        bx, by = axes.get_xlim(), axes.get_ylim()
        ratio = (bx[0] - bx[1]) / (by[0] - by[1])

        x = event.artist.get_xdata()
        y = event.artist.get_ydata()

        # points = filter(
        #     lambda e: e[0] in points_ind,
        #     enumerate(zip(x, y))
        # )
        points = enumerate(zip(x, y))

        mx = event.mouseevent.xdata
        my = event.mouseevent.ydata

        def dist_mouse(point):
            x, y = point[1]
            d = sqrt(
                    (((mx - x) / ratio) ** 2) +
                    ((my - y) ** 2)
            )
            return d

        closer = min(
            points, key=dist_mouse
        )

        return closer

    def _get_z_from_click(self, event):
        return event.ydata

    def _compute_hydrolics(self, z):
        profile = self.data

        points = profile.wet_points(z)
        station = profile._get_station(points)
        width = abs(station[0] - station[-1])

        poly = profile.wet_polygon(z)
        area = poly.area
        perimeter = poly.length

        return area, perimeter, width

    @timer
    def draw(self):
        self.init_axes()

        x = self.data.get_station()
        y = self.data.z()
        x_carto = self.data.x()
        y_carto = self.data.y()

        if (len(x_carto) < 3 or len(y_carto) < 3 or len(x) < 3):
            # Noting to do in this case
            return

        self.profile_line2D, = self.canvas.axes.plot(
            x, y, color=self.color_plot,
            lw=1.5, markersize=7, marker='+',
            picker=30
        )

        self.draw_annotation(x, y)
        self.draw_highligth()

        self.idle()

    def draw_annotation(self, x, y):
        gl = map(lambda p: p.name, self.data.points)

        # Add label on graph
        self.annotation = []
        for i, name in enumerate(list(gl)):
            annotation = self.canvas.axes.annotate(
                name, (x[i], y[i]),
                horizontalalignment='left',
                verticalalignment='top',
                annotation_clip=True,
                fontsize=10, color='black'
            )
            annotation.set_position((x[i], y[i]))
            annotation.set_color("black")
            self.annotation.append(annotation)

        al = 8.
        arrowprops = dict(
            clip_on=True,
            headwidth=5.,
            facecolor='k'
        )
        kwargs = dict(
            xycoords='axes fraction',
            textcoords='offset points',
            arrowprops=arrowprops,
        )

        self.canvas.axes.annotate("", (1, 0), xytext=(-al, 0), **kwargs)
        self.canvas.axes.annotate("", (0, 1), xytext=(0, -al), **kwargs)

        self.canvas.axes.spines[['top', 'right']].set_color('none')
        self.canvas.axes.yaxis.tick_left()
        self.canvas.axes.xaxis.tick_bottom()
        self.canvas.axes.set_facecolor('#F9F9F9')
        self.canvas.figure.patch.set_facecolor('white')

    def draw_highligth(self):
        points, hyd = self.highlight
        for p in self.hl_points:
            p[0].set_data([], [])

        self.hl_points = []

        for x, y in points:
            self.hl_points.append(
                self.canvas.axes.plot(
                    x, y,
                    color=self.color_plot_highlight,
                    lw=1.5, markersize=7, marker='+',
                    picker=30
                )
            )

        if hyd is not None:
            self.draw_highligth_z_line(*hyd)

    def draw_highligth_z_line(self, z, a, p, w):
        text = (
            f"Z = {z:.3f} m, " +
            f"{self._trad['width']} = {w:.3f} m,\n" +
            f"{self._trad['area']} = {a:.3f} m², " +
            f"{self._trad['perimeter']} = {p:.3f} m"
        )

        x = self.data.get_station()
        xlim = (x[0], x[-1])
        pos = (
            xlim[0] + (abs(xlim[0] - xlim[1]) * 0.05),
            z + 0.8
        )
        y = self.data.z()

        if self._z_note is None:
            self.draw_highligth_z_line_fill(x, y, z)

            self._z_line = self.canvas.axes.plot(
                xlim, [z, z],
                color=self.color_plot_river_water
            )

            self._z_note = self.canvas.axes.annotate(
                text, pos,
                horizontalalignment='left',
                verticalalignment='top',
                annotation_clip=True,
                color=self.color_plot_river_water,
                fontsize=9,
                fontweight='bold',
                alpha=0.7
            )
        else:
            self.draw_highligth_z_line_fill(x, y, z)

            self._z_line[0].set_data(xlim, [z, z])
            self._z_note.set_position(pos)
            self._z_note.set_text(text)

    def draw_highligth_z_line_fill(self, x, y, z):
        if self._z_fill_between is not None:
            self._z_fill_between.remove()

        self._z_fill_between = self.canvas.axes.fill_between(
            x, y, z,
            where=y <= z,
            facecolor=self.color_plot_river_water_zone,
            interpolate=True, alpha=0.7
        )

    @timer
    def update(self):
        self.draw_highligth()

        self.update_idle()
